/************************************************************************
File:		superlu_ccf.c  -- Column-Compressed-Format
Created:	10-23-02
Author:		G.Shi
Note:
	Copied from ../EXAMPLE/superlu.c
	Modified by G. Shi, 10-23-02, for testing some circuit systems.
	Data (A,b) are provided by files in the format of (row, col, value).
*************************************************************************/

/*
 * -- SuperLU routine (version 2.0) --
 * Univ. of California Berkeley, Xerox Palo Alto Research Center,
 * and Lawrence Berkeley National Lab.
 * November 15, 1997
 *
 */
#include "dsp_defs.h"
#include "util.h"
#include "time.h"

#define		HZ		CLK_TCK

double	Time()
{
	struct time {long  user, system, childuser, childsystem;} time;

	(void) times(&time);
	return  (double) time.user / (double) HZ;
}


static char *FileName;

int		m, n, nnz = 0;
double  *a, *rhs;
int     *rowind, *colptr; /* Used in Compressed Column Format (CCF). */


static int  ReadMatrixFromFile();
void		dPrint_Solution(char *what, SuperMatrix *A, int Len);


main(int argc, char *argv[])
{
/*
 * Purpose
 * =======
 * 
 * This is the small 5x5 example used in the Sections 1 and 2 of the 
 * User's Guide to illustrate how to call a SuperLU routine, and the
 * matrix data structures used by SuperLU.
 *
 */

    SuperMatrix A, L, U, B;
    double   s, u, p, e, r, l;
    int      *perm_r; /* row permutations from partial pivoting */
    int      *perm_c; /* column permutation vector */
    int      nrhs = 1, info, i, permc_spec;

	/* Newly added variables */
	int		Error;
	double 	StartTime, SolveTime, OrderTime, Time0, TotalTime;


    /* Get the beginning time for measuring the time duration. */
    Time0 = Time(); 

	/* Read matrix file. */
	if (argc == 1) {
		fprintf(stderr, "Error: No data file.\nUsage: superlu data_file.\n");
		exit(1);
	}
	else 
		FileName = *(++argv);

	Error = ReadMatrixFromFile();

	if (Error) {
		fprintf(stderr, "Error: Failed reading data file. \n ");
		exit(1);
	}

	/* Print to test */
	/*
	printf("\n at: \n");
	for (i = 0; i < nnz; i++)   printf(" %.2lf ", at[i]);
	printf("\n rowind : \n");
	for (i = 0; i < nnz; i++)   printf(" %d ", rowind[i]);
	printf("\n colptr: \n");
	for (i = 0; i < n+1; i++)   printf(" %d ", colptr[i]);
	*/
	printf("\n");

	/* Terminate the program here. */
	/*printf("\n"); exit(1);*/ 

    /* Create matrix A in the format expected by SuperLU. */
    dCreate_CompCol_Matrix(&A, m, n, nnz, a, rowind, colptr, NC, D_, GE);
    
    /* Create right-hand side matrix B. */
    dCreate_Dense_Matrix(&B, m, nrhs, rhs, m, DN, D_, GE);

    if ( !(perm_r = intMalloc(m)) ) ABORT("Malloc fails for perm_r[].");
    if ( !(perm_c = intMalloc(n)) ) ABORT("Malloc fails for perm_c[].");

    /*
     * Get column permutation vector perm_c[], according to permc_spec:
     *   permc_spec = 0: use the natural ordering 
     *   permc_spec = 1: use minimum degree ordering on structure of A'*A
     *   permc_spec = 2: use minimum degree ordering on structure of A'+A
     */    	
    /*permc_spec = 1;*/   /* Options: 0, 1, 2 */

	/* Record time before keyboard input. */
	TotalTime = Time() - Time0;

	do {
		printf("Choose ordering option (0 -- None; 1 -- A'*A; 2 -- A'+A) : ");
		scanf("%d", &permc_spec);
	} while ( permc_spec != 0 && permc_spec != 1 && permc_spec != 2 );

	Time0 = Time();

	putchar('\n');
	/*printf("permc_spec = %d\n", permc_spec);*/

	StartTime = Time();
    get_perm_c(permc_spec, &A, perm_c);
	OrderTime = Time() - StartTime;

	fprintf(stderr, "Pre-ordering done.\n");
	fflush(stderr);

	/* Solve and record time. */
	printf("Start solving ... \n");
	fflush(stderr);
    
	StartTime = Time();
    dgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
	SolveTime = Time() - StartTime;

	/* Check the 'info'. */
    printf("\ninfo = %d ", info);   /* info = 0 : successful exit */
    if (info == 0)  printf(": Successful exit. \n");
    else            printf(": Unsuccessful exit. \n");

	/* Print the times used. */
	putchar('\n');
	printf("* Order time = %.3lf seconds.\n", OrderTime);
	printf("* Factor and Solve time = %.3lf seconds (Option %d).\n", \
			SolveTime, permc_spec);

	TotalTime += Time() - Time0;
    printf("* Total time used = %.3lf seconds.\n", TotalTime);

	/* Print the solution which is in 'B'. */
	dPrint_Solution("Truncated Solution:", &B, 10);
	
	/*
    dPrint_Dense_Matrix("B", &B);
    dPrint_CompCol_Matrix("A", &A);
    dPrint_CompCol_Matrix("U", &U);
    dPrint_SuperNode_Matrix("L", &L);
    print_int_vec("\nperm_r", m, perm_r);
	*/	

    /* De-allocate storage */
    SUPERLU_FREE (a);
    SUPERLU_FREE (rowind);
    SUPERLU_FREE (colptr);
	/* Above three lines newly added. */

    SUPERLU_FREE (rhs);
    SUPERLU_FREE (perm_r);
    SUPERLU_FREE (perm_c);
    Destroy_CompCol_Matrix(&A);
    Destroy_SuperMatrix_Store(&B);
    Destroy_SuperNode_Matrix(&L);
    Destroy_CompCol_Matrix(&U);
}


static int  ReadMatrixFromFile()
{
	FILE  	*pMatrixFile;
	int   	Error, Row, Col, Count = 0, LineNumber, RHS_Col;
	int	  	Size = 0;
	long  	Reads;
	double 	Real;
	char  	sInput[BUFSIZ], sType[BUFSIZ];
	static char Message[BUFSIZ];

	static char *EndOfFile = "Unexpected end of file '%s' at line %d.\n";
	static char *Syntax = "Syntax error in file '%s' at line %d.\n";
	static char *Index = "Index error in file '%s' at line %d.\n";

	int		i, j;    
	/* Indices used to trace the three arrays in CCF. */
	int		curr_col;		/* The current column index. */
	int		M;				/* For printing test. */

	pMatrixFile = fopen(FileName, "r");
	if (pMatrixFile == NULL) {
		fprintf(stderr, "Failed to open file: %s \n", FileName);
		return 1;
	}
	
	LineNumber = 1;

	/* Read and print label. */
	if (NULL == fgets(Message, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}

	/* Check by print */
	/* printf(" Message = %s \n", Message); */

	/* Read the size of matrix. */
	LineNumber++;

	if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}
	if ((Reads = sscanf(sInput, "%d %s", &Size, sType)) < 1) {
		fprintf(stderr, Syntax, FileName, LineNumber);
		return 1;
	}

	/* Print to Check */
	/* printf(" Reads = %d \n", Reads); */

	/* Set the size of matrix A. (m, n) are globals. */
	m = n = Size;

	/* Read data file once to get the nonzero count (nnz). */
	nnz = 0;	/* Initialize num_of_nonzeros */
	do {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);

		if (Reads != 3) {
			fprintf(stderr, Syntax, FileName, LineNumber);
			return 1;
		}
		if (Row < 0 || Row > Size || Col < 0 || Col > Size) {
			fprintf(stderr, Index, FileName, LineNumber);
			return 1;
		}

		/* Print to check */
		/* printf(" %d, %d, %.2lf \n", Row, Col, Real); */

		/* Get the total num of nonzeros so that memory can be allocated. */
		if (Row > 0 && Col > 0)		++nnz;

	} while (Row != 0 && Col != 0);

	/* Print to check */
	printf("\nSize = %d,  Type = %s \n", Size, sType);
	printf("nnz = %d \n", nnz); 
	printf("sparsity (nnz/n) = %.6lf \n", (double) nnz/Size ); 

	/* Allocate memory for Column Compressed Format (CCF). */

    if ( !(a = doubleMalloc(nnz)) ) ABORT("Malloc fails for a[].");
    if ( !(rowind = intMalloc(nnz)) ) ABORT("Malloc fails for colind[].");
    if ( !(colptr = intMalloc(n+1)) ) ABORT("Malloc fails for rowptr[].");

	/* Allocate memory for RHS. */

    if ( !(rhs = doubleMalloc(m)) ) ABORT("Malloc fails for rhs[].");


    /* Read data file again to set up arrays in Column Compressed Format. */

	/* Resume the file pointer */
	rewind(pMatrixFile);

	/* Skip the first two lines. */
	fgets(Message, BUFSIZ, pMatrixFile);
	fgets(Message, BUFSIZ, pMatrixFile);
	LineNumber = 2;

	LineNumber++;
    fgets(sInput, BUFSIZ, pMatrixFile);
    Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);

	i = j = 0;
	curr_col = 0;		/* col index counts from 1 in the data file */

    while (Row > 0 && Col > 0) {

		a[i] = Real;			/* nonzero value */
		rowind[i] = Row - 1;	/* rowind = row index */

		if (Col > curr_col) {
			colptr[j++] = i; 
			curr_col++;
		}
		i++;

		LineNumber++;
        fgets(sInput, BUFSIZ, pMatrixFile);
        Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);
    } 

	colptr[j] = i;		/* Append the ending index in rowptr. */

	/* Print to test */
	/*	
	M = (nnz > 20)? 20 : nnz;
	printf("\na: \n");
	for (i = 0; i < M; i++)   printf(" %lf ", a[i]);
	printf("\nrowind : \n");
	for (i = 0; i < M; i++)   printf(" %d ", rowind[i]);
	printf("\ncolptr: \n");
	for (i = 0; i < M; i++)   printf(" %d ", colptr[i]);
	*/	

	/* Continue to set up the RHS */
	for (i = 0; i < Size; i++) {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		Reads = sscanf(sInput, "%lf", &Real);

		if (Reads != 1) {
			fprintf(stderr, Syntax, FileName, LineNumber);
			return 1;
		}

		rhs[i] = Real;
	}

	/* Print to test */
	/*
	printf("\n rhs: \n");
	for (i = 0; i < m; i++)   printf(" %.2lf ", rhs[i]);
	*/


	(void) fclose(pMatrixFile);

	printf("\n");

	return 0;		/* No error. */
}

void	dPrint_Solution(char *what, SuperMatrix *S, int Len)
{
    DNformat     *Astore;
    register int i;
    double       *dp;

    printf("\n%s\n", what);
    Astore = (DNformat *) S->Store;
    dp = (double *) Astore->nzval;
    for (i = 0; i < Len; ++i) 
		printf("%+16.9lg \n", dp[i]);
    printf("\n");
    fflush(stdout);
}

